#pragma once

#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>

namespace AMDiS
{
  template <class WorldVector, class Range>
    template <class RowBasis>
  void DirichletBC<WorldVector, Range>::initImpl(
      RowBasis const& rowBasis, Dune::TypeTree::LeafNodeTag)
  {
    using Dune::Functions::forEachBoundaryDOF;
    using LV = typename RowBasis::LocalView;
    using I = typename RowBasis::GridView::Intersection;
    dirichletNodes_.resize(rowBasis.dimension());
    forEachBoundaryDOF(rowBasis, [&](int localIndex, LV const& localView, I const& intersection) {
      dirichletNodes_[localView.index(localIndex)] = predicate_(intersection.geometry().center());
    });
  }

  template <class WorldVector, class Range>
    template <class RowBasis>
  void DirichletBC<WorldVector, Range>::initImpl(
      RowBasis const& rowBasis, Dune::TypeTree::PowerNodeTag)
  {
    using Dune::Functions::subspaceBasis;
    using Node = typename RowBasis::LocalView::Tree;
    using ChildNode = typename Node::template Child<0>::type;

    auto tp = rowBasis.prefixPath();
    auto const& basis = rowBasis.rootBasis();
    for (std::size_t i = 0; i < degree(rowBasis.localView().tree()); ++i)
      initImpl(subspaceBasis(basis, push_back(tp,i)), typename ChildNode::NodeTag{});
  }

  template <class WorldVector, class Range>
    template <class RowBasis>
  void DirichletBC<WorldVector, Range>::initImpl(
      RowBasis const& rowBasis, Dune::TypeTree::CompositeNodeTag)
  {
    using Dune::Functions::subspaceBasis;
    using Node = typename RowBasis::LocalView::Tree;
    auto tp = rowBasis.prefixPath();
    auto const& basis = rowBasis.rootBasis();
    forEach(range_<0,Node::CHILDREN>, [&,this](auto const _i) {
      using ChildNode = typename Node::template Child<decltype(_i)::value>::type;
      this->initImpl(subspaceBasis(basis, push_back(tp,_i)), typename ChildNode::NodeTag{});
    });
  }

} // end namespace AMDiS
