#pragma once

#include <dune/istl/preconditioner.hh>

#include <amdis/linear_algebra/RunnerInterface.hpp>
#include <amdis/linear_algebra/SolverInfo.hpp>
#include <amdis/linear_algebra/eigen/PreconConfig.hpp>
#include <amdis/linear_algebra/eigen/SolverConfig.hpp>

namespace AMDiS
{
  template <class IterativeSolver, class Matrix, class VectorX, class VectorB>
  class IterativeRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
    using SolverCfg = SolverConfig<IterativeSolver>;
    using PreconCfg = PreconConfig<typename IterativeSolver::Preconditioner>;

  public:
    IterativeRunner(std::string prefix)
      : solver_{}
    {
      SolverCfg::init(prefix, solver_);
      PreconCfg::init(prefix + "->precon", solver_.preconditioner());
    }


    /// Implementes \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      solver_.compute(A);

      test_exit(solver_.info() == Eigen::Success,
        "Error in solver.compute(matrix)");
    }

    /// Implementes \ref RunnerInterface::exit()
    virtual void exit() override {}

    /// Implementes \ref RunnerInterface::solve()
    virtual int solve(Matrix const& A, VectorX& x, VectorB const& b,
                      SolverInfo& solverInfo) override
    {
      solver_.setTolerance(solverInfo.getRelTolerance());
      x = solver_.solveWithGuess(b, x);

      auto r = VectorB(b);
      if (x.norm() != 0)
        r -= A * x;

      solverInfo.setAbsResidual(r.norm());
      solverInfo.setRelResidual(solver_.error());
      solverInfo.setError(solver_.info());
      msg("number of iteration: {}", solver_.iterations());

      return solver_.info() == Eigen::Success ? 0 : int(solver_.info());
    }

  private:
    IterativeSolver solver_;
  };

} // end namespace AMDiS
